from __future__ import annotations

import functools
import logging
from contextlib import asynccontextmanager
from inspect import iscoroutinefunction, signature
from typing import Callable, Dict, List, Optional, Tuple, Type, final

from icij_common.pydantic_utils import ICIJModel, ICIJSettings
from icij_worker.namespacing import Namespacing
from icij_worker.typing_ import Dependency
from icij_worker.utils import run_deps
from icij_worker.utils.imports import import_variable

logger = logging.getLogger(__name__)

PROGRESS_HANDLER_ARG = "progress"


class AsyncAppConfig(ICIJSettings):
    late_ack: bool = False
    max_task_queue_size: Optional[int] = None

    class Config:
        env_prefix = "ICIJ_APP_"


class RegisteredTask(ICIJModel):
    task: Callable
    recover_from: Tuple[Type[Exception], ...] = tuple()
    max_retries: Optional[int]
    namespace: Optional[str]


class AsyncApp:
    def __init__(
        self,
        name: str,
        config: AsyncAppConfig = None,
        dependencies: Optional[List[Dependency]] = None,
        namespacing: Optional[Namespacing] = None,
    ):
        self._name = name
        if config is None:
            config = AsyncAppConfig()  # This will load from the env
        self._config = config
        self._registry = dict()
        if dependencies is None:
            dependencies = []
        self._dependencies = dependencies
        if namespacing is None:
            namespacing = Namespacing()
        self._namespacing = namespacing

    @property
    def config(self) -> AsyncAppConfig:
        return self._config

    def with_config(self, value: AsyncAppConfig) -> AsyncApp:
        if not isinstance(value, AsyncAppConfig):
            raise TypeError(f"Expected {AsyncAppConfig.__name__}, got {value}")
        self._config = value
        return self

    @property
    def registry(self) -> Dict[str, RegisteredTask]:
        return self._registry

    @property
    def registered_keys(self) -> List[str]:
        return sorted(self._registry)

    @functools.cached_property
    def name(self) -> str:
        return self._name

    @functools.cached_property
    def namespacing(self) -> Namespacing:
        return self._namespacing

    def with_namespacing(self, ns: Namespacing) -> AsyncApp:
        self._namespacing = ns
        return self

    def task(
        self,
        name: Optional[str] = None,
        recover_from: Tuple[Type[Exception]] = tuple(),
        max_retries: Optional[int] = None,
        *,
        namespace: Optional[str] = None,
    ) -> Callable:
        if callable(name) and not recover_from and max_retries is None:
            f = name
            return functools.partial(
                self._register_task, name=f.__name__, namespace=namespace
            )(f)
        if max_retries is None:
            max_retries = 3
        return functools.partial(
            self._register_task,
            name=name,
            recover_from=recover_from,
            max_retries=max_retries,
            namespace=namespace,
        )

    @final
    @asynccontextmanager
    async def lifetime_dependencies(self, **kwargs):
        ctx = f"{self.name} async app"
        async with run_deps(self._dependencies, ctx=ctx, **kwargs):
            yield

    def _register_task(
        self,
        f: Callable,
        *,
        name: Optional[str] = None,
        recover_from: Tuple[Type[Exception]] = tuple(),
        max_retries: Optional[int] = None,
        namespace: Optional[str],
    ) -> Callable:
        if not iscoroutinefunction(f) and supports_progress(f):
            msg = (
                f"{f} is not a coroutine, progress is not supported as progress"
                f" reporting is inherently async, turn your function task into a"
                f" coroutine if necessary and use `await progress(my_progress)`"
            )
            raise ValueError(msg)
        if name is None:
            name = f.__name__
        registered = self._registry.get(name)
        if registered is not None:
            raise ValueError(f'Task "{name}" is already registered: {registered}')
        self._registry[name] = RegisteredTask(
            task=f,
            max_retries=max_retries,
            recover_from=recover_from,
            namespace=namespace,
        )

        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapped

    @classmethod
    def load(cls, app_path: str, config: Optional[AsyncAppConfig] = None) -> AsyncApp:
        app = import_variable(app_path)
        if config is not None:
            app.with_config(config)
        return app

    def filter_tasks(self, namespace: str) -> AsyncApp:
        kept = {
            t_name
            for t_name, t in self._registry.items()
            if self._namespacing.app_tasks_filter(
                task_namespace=t.namespace, app_namespace=namespace
            )
        }
        discarded = set(self._registry) - kept
        logger.info(
            "Applied namespace filtering:\n- running: %s\n- discarded: %s",
            ", ".join(sorted(kept)),
            ", ".join(sorted(discarded)),
        )
        self._registry = {k: self._registry[k] for k in kept}
        return self


def supports_progress(task_fn) -> bool:
    return any(
        param.name == PROGRESS_HANDLER_ARG
        for param in signature(task_fn).parameters.values()
    )



from __future__ import annotations

import logging
from datetime import datetime
from functools import cached_property
from typing import Dict, List, Optional, TypeVar, Union, cast

from aio_pika import connect_robust
from aio_pika.abc import AbstractExchange, AbstractQueueIterator
from aiormq import DeliveryError

from icij_worker import AsyncApp, Task, TaskManager
from icij_worker.event_publisher.amqp import RobustConnection
from icij_worker.exceptions import TaskQueueIsFull
from icij_worker.namespacing import Routing
from icij_worker.objects import (
    CancelEvent,
    ErrorEvent,
    ManagerEvent,
    Message,
    ResultEvent,
    TaskState,
)
from icij_worker.task_storage import TaskStorage
from icij_worker.utils.amqp import AMQPMixin

S = TypeVar("S", bound=TaskStorage)

logger = logging.getLogger(__name__)


class AMQPTaskManager(TaskManager, AMQPMixin):

    def __init__(
        self,
        app: AsyncApp,
        task_store: TaskStorage,
        *,
        broker_url: str,
        connection_timeout_s: float = 1.0,
        reconnection_wait_s: float = 5.0,
        inactive_after_s: Optional[float] = None,
    ):
        super().__init__(app)
        super(TaskManager, self).__init__(
            broker_url,
            connection_timeout_s=connection_timeout_s,
            reconnection_wait_s=reconnection_wait_s,
            inactive_after_s=inactive_after_s,
        )
        self._storage = task_store

        self._loops = set()

        self._task_x: Optional[AbstractExchange] = None
        self._worker_evt_x: Optional[AbstractExchange] = None
        self._manager_messages_it: Optional[AbstractQueueIterator] = None

        self._task_namespaces: Dict[str, Optional[str]] = dict()

    async def _aenter__(self) -> AMQPTaskManager:
        logger.info("starting task manager connection workflow...")
        await self._exit_stack.__aenter__()  # pylint: disable=unnecessary-dunder-call
        await self._exit_stack.enter_async_context(self._storage)
        await self._connection_workflow()
        await self._ensure_task_queues()
        self._manager_messages_it = (
            await self._get_queue_iterator(
                self.manager_evt_routing(), declare_exchanges=False
            )
        )[0]
        logger.info("starting consume loops..")
        return self

    async def _aexit__(self, exc_type, exc_val, exc_tb):
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)

    @cached_property
    def _app_id(self) -> str:
        return self.app_name

    async def get_task(self, task_id: str) -> Task:
        return await self._storage.get_task(task_id)

    async def get_task_namespace(self, task_id: str) -> Optional[str]:
        return await self._storage.get_task_namespace(task_id)

    async def get_task_errors(self, task_id: str) -> List[ErrorEvent]:
        return await self._storage.get_task_errors(task_id)

    async def get_task_result(self, task_id: str) -> ResultEvent:
        return await self._storage.get_task_result(task_id)

    async def get_tasks(
        self,
        namespace: Optional[str],
        *,
        task_name: Optional[str] = None,
        state: Optional[Union[List[TaskState], TaskState]] = None,
        **kwargs,
    ) -> List[Task]:
        return await self._storage.get_tasks(
            namespace, task_name=task_name, state=state
        )

    async def save_task_(self, task: Task, namespace: Optional[str]) -> bool:
        return await self._storage.save_task_(task, namespace)

    async def save_result(self, result: ResultEvent):
        await self._storage.save_result(result)

    async def save_error(self, error: ErrorEvent):
        await self._storage.save_error(error)

    async def _consume(self) -> ManagerEvent:
        msg = await self._manager_messages_it.__anext__()
        await msg.ack()
        return cast(ManagerEvent, Message.parse_raw(msg.body))

    async def _enqueue(self, task: Task):
        namespace = await self._storage.get_task_namespace(task.id)
        await self._ensure_task_queues()
        routing = self._namespacing.amqp_task_routing(namespace)
        try:
            await self._publish_message(
                task,
                exchange=self._task_x,
                routing_key=routing.routing_key,
                mandatory=True,  # This is important
            )
        except DeliveryError as e:
            raise TaskQueueIsFull(self.max_task_queue_size) from e

    async def _requeue(self, task: Task):
        namespace = await self._storage.get_task_namespace(task.id)
        routing = self._namespacing.amqp_task_routing(namespace)
        try:
            await self._publish_message(
                task,
                exchange=self._task_x,
                routing_key=routing.routing_key,
                mandatory=True,  # This is important
            )
        except DeliveryError as e:
            raise TaskQueueIsFull(self.max_task_queue_size) from e

    async def cancel(self, task_id: str, *, requeue: bool):
        cancel_event = CancelEvent(
            task_id=task_id, requeue=requeue, created_at=datetime.now()
        )
        # TODO: for now cancellation is not namespaced, workers from other namespace
        #  are responsible to ignoring the broadcast. That could be easily implemented
        #  in the future but will need sync with Java
        routing = self.worker_evt_routing().routing_key
        await self._publish_message(
            cancel_event,
            exchange=self._worker_evt_x,
            routing_key=routing,
            mandatory=True,  # This is important
        )

    async def _connection_workflow(self):
        logger.debug("creating connection...")
        self._connection_ = await connect_robust(
            self._broker_url,
            timeout=self._connection_timeout_s,
            reconnect_interval=self._reconnection_wait_s,
            connection_class=RobustConnection,
        )
        await self._exit_stack.enter_async_context(self._connection)
        logger.debug("creating channel...")
        self._channel_ = await self._connection.channel(
            publisher_confirms=True, on_return_raises=False
        )
        await self._exit_stack.enter_async_context(self._channel)
        await self._channel.set_qos(prefetch_count=1, global_=True)
        logger.info("channel opened !")
        task_routing = self.default_task_routing()
        logger.debug("(re)declaring routing %s...", task_routing)
        task_queue_args = None
        if self.max_task_queue_size is not None:
            task_queue_args = {
                "x-overflow": "reject-publish",
                "x-max-length": self.max_task_queue_size,
            }
        await self._create_routing(
            task_routing,
            declare_exchanges=True,
            declare_queues=True,
            durable_queues=True,
            queue_args=task_queue_args,
        )
        for routing in self._other_routings:
            logger.debug("(re)declaring routing %s...", routing)
            await self._create_routing(
                routing,
                declare_exchanges=True,
                declare_queues=True,
                durable_queues=True,
            )
        await self._create_routing(
            AMQPMixin.worker_evt_routing(), declare_exchanges=True, declare_queues=False
        )
        self._task_x = await self._channel.get_exchange(
            self.default_task_routing().exchange.name, ensure=True
        )
        self._manager_evt_x = await self._channel.get_exchange(
            self.manager_evt_routing().exchange.name, ensure=True
        )
        self._worker_evt_x = await self._channel.get_exchange(
            self.worker_evt_routing().exchange.name, ensure=True
        )
        logger.info("connection workflow complete")

    @cached_property
    def _other_routings(self) -> List[Routing]:
        worker_events_routing = AMQPMixin.worker_evt_routing()
        manager_events_routing = AMQPMixin.manager_evt_routing()
        return [worker_events_routing, manager_events_routing]

    async def _ensure_task_queues(self):
        supported_ns = set(t.namespace for t in self._app.registry.values())
        for namespace in supported_ns:
            routing = self._namespacing.amqp_task_routing(namespace)
            task_routing = self.default_task_routing()
            dlx_name = task_routing.dead_letter_routing.exchange.name
            dl_routing_key = task_routing.dead_letter_routing.routing_key
            arguments = {
                "x-dead-letter-exchange": dlx_name,
                "x-dead-letter-routing-key": dl_routing_key,
            }
            if self.max_task_queue_size is not None:
                arguments["x-max-length"] = self.max_task_queue_size
                arguments["x-overflow"] = "reject-publish"
            queue = await self._channel.declare_queue(
                routing.queue_name, durable=True, arguments=arguments
            )
            await queue.bind(routing.exchange.name, routing.routing_key)



from __future__ import annotations

from abc import ABC
from pathlib import Path
from typing import ClassVar, Optional, Union

from pydantic import Field, Protocol
from pydantic.parse import load_file

from icij_worker.utils.registrable import RegistrableConfig


class WorkerConfig(RegistrableConfig, ABC):
    registry_key: ClassVar[str] = Field(const=True, default="type")

    # TODO: is app_dependencies_path better ?
    app_bootstrap_config_path: Optional[Path] = None
    inactive_after_s: Optional[float] = None
    log_level: str = "INFO"
    task_queue_poll_interval_s: float = 1.0
    type: ClassVar[str]

    class Config:
        env_prefix = "ICIJ_WORKER_"

    @classmethod
    def parse_file(
        cls: WorkerConfig,
        path: Union[str, Path],
        *,
        content_type: str = None,
        encoding: str = "utf8",
        proto: Protocol = None,
        allow_pickle: bool = False,
    ) -> WorkerConfig:
        obj = load_file(
            path,
            proto=proto,
            content_type=content_type,
            encoding=encoding,
            allow_pickle=allow_pickle,
            json_loads=cls.__config__.json_loads,
        )
        worker_type = obj.pop(WorkerConfig.registry_key.default)
        subcls = WorkerConfig.resolve_class_name(worker_type)
        return subcls(**obj)



from enum import Enum, unique

from .config import WorkerConfig
from .worker import Worker


@unique
class WorkerType(str, Enum):
    # pylint: disable=invalid-name@
    mock = "mock"
    neo4j = "neo4j"
    amqp = "amqp"



from .app import AsyncApp
from .objects import Task, TaskError, ManagerEvent, ResultEvent, TaskState, Message
from .task_manager import TaskManager
from .worker import Worker, WorkerConfig, WorkerType
from .namespacing import Namespacing

try:
    from icij_worker.worker.amqp import AMQPWorker, AMQPWorkerConfig
    from icij_worker.event_publisher.amqp import AMQPPublisher
    from icij_worker.task_manager.amqp import AMQPTaskManager
except ImportError:
    pass

try:
    from icij_worker.worker.neo4j_ import Neo4jWorker, Neo4jWorkerConfig
    from icij_worker.event_publisher.neo4j_ import Neo4jEventPublisher
    from icij_worker.task_manager.neo4j_ import Neo4JTaskManager
except ImportError:
    pass

try:
    from icij_worker.task_storage.fs import FSKeyValueStorage
except ImportError:
    pass

from .backend import WorkerBackend
from .event_publisher import EventPublisher



AMQP_MANAGER_EVENTS_X = "exchangeManagerEvents"
AMQP_MANAGER_EVENTS_QUEUE = "MANAGER_EVENT"
AMQP_MANAGER_EVENTS_ROUTING_KEY = "routingKeyManagerEvents"

AMQP_MANAGER_EVENTS_DL_X = "exchangeDLQManagerEvents"
AMQP_MANAGER_EVENTS_DL_QUEUE = "MANAGER_EVENT_DLQ"
AMQP_MANAGER_EVENTS_DL_ROUTING_KEY = "routingKeyDLQManagerEvents"

AMQP_TASKS_DL_X = "exchangeDLQTasks"
AMQP_TASKS_DL_QUEUE = "TASK_DLQ"
AMQP_TASKS_DL_ROUTING_KEY = "routingKeyDLQTasks"

AMQP_TASKS_X = "exchangeMainTasks"
AMQP_TASKS_QUEUE = "TASK"
AMQP_TASKS_ROUTING_KEY = "routingKeyMainTasks"

AMQP_WORKER_EVENTS_X = "exchangeWorkerEvents"
AMQP_WORKER_EVENTS_QUEUE = "WORKER_EVENT"
AMQP_WORKER_EVENTS_ROUTING_KEY = "routingKeyWorkerEvents"



from abc import ABC, abstractmethod
from functools import cached_property
from typing import Type, TypeVar

from pydantic import BaseSettings

T = TypeVar("T", bound="FromConfig")
C = TypeVar("C", bound=BaseSettings)


class FromConfig(ABC):
    @classmethod
    @abstractmethod
    def _from_config(cls: Type[T], config: C, **extras) -> T: ...

    @abstractmethod
    def _to_config(self) -> C: ...

    @cached_property
    def config(self) -> C:
        return self._to_config()
